#!/usr/bin/python

from pylab import *
from sympy import var
import numpy as np
from scipy.optimize import newton as newtonSP
from sys import argv
from exampleproblems import *

def solveTSVD(A,b,normNoise):
    """
    Return the solution x_tsvd corresponding to the truncated singular value decomposition solution.
    """
    [U,S,V] = np.linalg.svd(A)
    V = V.T
    m,n = np.shape(A)
    w   = np.dot(U.T, b)      
    for j in xrange(0,n):
        y = np.zeros(n)
        for k in xrange(0,j+1):
            y[k] = w[k] / S[k]
        norm = np.linalg.norm(np.dot(np.diag(S), y) - w,2)**2
        if norm < normNoise**2:
            break
    x = np.dot(V,y)
    return x

def solveTSVDAlternate(A,b,normNoise):
    """
    Return the solution x_tsvd corresponding to the truncated singular value decomposition solution.
    """
    [U,S,V] = np.linalg.svd(A)
    V = V.T
    m,n = np.shape(A)
    w   = np.dot(U.T, b)      
    for j in xrange(0,n):
        y = np.zeros(n)
        for k in xrange(0,j+1):
            y[k] = w[k] / S[k]
        norm = np.linalg.norm(np.dot(np.diag(S), y) - w,2)**2
        if norm < normNoise**2:
            break
    mu = tikhonovMu(S, w, normNoise)
    for k in xrange(j+1,n):
        y[k] = w[k] / mu
    x = np.dot(V,y)
    return x

def optimalMu(A,bt,x,normNoise):
    n,m = A.shape
    [U,S,V] = np.linalg.svd(A)
    V = V.T
    b = tikhonovMu(S,bt,normNoise)
    a = 0
    Sfull = np.diag(S)
    minResidual  = 10000
    minMu = b

    for mu in np.arange(a,b,(b-a)/100):
        Snew  = np.zeros((n,n))
        for j in range(0,n):
            Snew[j,j] = 1/(S[j]**2 + mu)
        xMu = np.dot(V, Snew)
        xMu = np.dot(xMu, Sfull)
        xMu = np.dot(xMu, bt)
        f   = np.linalg.norm(xMu - x,2)

        if f < minResidual:
            minResidual  = f
            minMu = mu
    return (minMu, minResidual)

def func(f,mu):
    return lambda x: f.subs({mu: x})

def fprime(S,b,nu):
    fprime = 0
    n, = np.shape(S)
    for j in range(0,n):
        fprime += -2*b[j]**2 * S[j]**2 / (1 + nu* S[j]**2)**3
    return fprime

def f(S,b,normNoise,nu):
    f = 0
    n, = np.shape(S)
    for j in range(0,n):
        f += b[j]**2 / (1 + nu * S[j]**2)**2
    f -= normNoise**2
    return f

def fRGprime(S,b,nu):
    fprime = 0
    n, = np.shape(S)
    for j in range(0,n):
        fprime += -3*b[j]**2 * S[j]**2 / (1 + nu * S[j]**2)**4
    return fprime

def fRG(S,b,normNoise,nu):
    f = 0
    n, = np.shape(S)
    for j in range(0,n):
        f += b[j]**2 / (1 + nu* S[j]**2)**3
    f -= normNoise**2
    return f

def newton(S,b,normNoise):
    t     = -1
    tnext = 1e-5
    while (tnext > t):
        t     = tnext
        tnext = t - f(S,b,normNoise,t) / fprime(S,b,t)
    return np.abs(1. / tnext)

def newtonRG(S,b,normNoise):
    t     = -1
    tnext = 1e-5
    while (tnext > t):
        t     = tnext
        tnext = t - fRG(S,b,normNoise,t) / fRGprime(S,b,t)
    return 1. / tnext

def tikhonovMuRG(S,b,normNoise):
    """
    Returns the solution vector x_mu corresponding to the Tikhonov regularazation parameter
    using RG.
    """
    mu = newtonRG(S,b,normNoise)
    return mu

def tikhonovMu(S,b,normNoise):
    """
    Returns a solution for the regularization parameter based on regular Tikhonov regularization.
    """
    mu = newton(S,b,normNoise)
    return mu

def sigmaNewAlternate(S,mu):
    """
    Returns the new sigma with mu as it's value along the diagonal after S(j)**2 < mu.
    Also, if S(j)**2 < mu/2, it'll set S(j) = 0.
    """
    n, = np.shape(S)
    S = S**2
    for j in range(0,n):
        if S[j] < mu and 2*S[j] > mu:
            S[j] = mu
        elif 2*S[j] < mu:
            S[j] = 0
    return np.diag(S[:])

def sigmaNew(S,mu):
    """
    Returns the new sigma with mu as it's value along the diagonal after S(j) < mu.
    """
    n, = np.shape(S)
    S = S**2
    for j in range(0,n):
        if S[j] < mu:
            S[j] = mu
    return np.diag(S[:])

def solveTikhonovRG(A,b,normNoise):
    """
    Returns the solution vector x_mu corresponding to the Tikhonov using the new RG method.
    """
    [U,S,V] = np.linalg.svd(A)
    V = V.T
    m,n   = np.shape(A)
    bt    = np.dot(U.T, b)
    mu    = tikhonovMuRG(S,bt,normNoise)
    Sfull = np.diag(S)
    Snew  = np.zeros((n,n))
    for j in range(0,n):
        Snew[j,j] = 1/(S[j]**2 + mu)
    xMuTik = np.dot(V, Snew)
    xMuTik = np.dot(xMuTik, Sfull)
    xMuTik = np.dot(xMuTik, bt)
    return xMuTik

def solveTikhonov(A,b,normNoise):
    """
    Returns the solution vector x_mu corresponding to the Tikhonov regularization parameter.
    """
    [U,S,V] = np.linalg.svd(A)
    V = V.T
    m,n   = np.shape(A)
    bt    = np.dot(U.T, b)
    mu    = tikhonovMu(S,bt,normNoise)
    Sfull = np.diag(S)
    Snew  = np.zeros((n,n))
    for j in range(0,n):
        Snew[j,j] = 1/(S[j]**2 + mu)
    xMuTik = np.dot(V, Snew)
    xMuTik = np.dot(xMuTik, Sfull)
    xMuTik = np.dot(xMuTik, bt)
    return xMuTik

def solveTikhonovOptimal(A,b,x,normNoise):
    """
    Returns the solution vector x_mu corresponding to the Tikhonov regularization parameter.
    """
    [U,S,V] = np.linalg.svd(A)
    V = V.T
    m,n   = np.shape(A)
    bt    = np.dot(U.T, b)
    mu, empty   = optimalMu(A,bt,x,normNoise)
    Sfull = np.diag(S)
    Snew  = np.zeros((n,n))
    for j in range(0,n):
        Snew[j,j] = 1/(S[j]**2 + mu)
    xMuTik = np.dot(V, Snew)
    xMuTik = np.dot(xMuTik, Sfull)
    xMuTik = np.dot(xMuTik, bt)
    return xMuTik

def solveLotharRG(A,b,normNoise):
    """
    Returns the solution for x based on Lothar's new method using the RG method of determining mu.
    """
    [U,S,V] = np.linalg.svd(A)
    V = V.T
    m,n   = np.shape(A)
    bt    = np.dot(U.T, b)
    mu    = tikhonovMuRG(S,bt,normNoise)
    Snew  = sigmaNew(S,mu)
    Sfull = np.diag(S)
    for j in range(0,n):
        if Snew[j,j] != 0:
            Snew[j,j] = 1 / Snew[j,j]
    xMu = np.dot(V,Snew)
    xMu = np.dot(xMu,Sfull)
    xMu = np.dot(xMu,bt)
    return xMu

def solveLotharAlternate(A,b,normNoise):
    """
    Returns the solution for x based on Lothar's new method by truncating sigma after sigma closer to 0 than mu.
    """
    [U,S,V] = np.linalg.svd(A)
    V = V.T
    m,n   = np.shape(A)
    bt    = np.dot(U.T, b)
    mu    = tikhonovMu(S,bt,normNoise)
    Snew  = sigmaNewAlternate(S,mu)
    Sfull = np.diag(S)
    for j in range(0,n):
        if Snew[j,j] != 0:
            Snew[j,j] = 1 / Snew[j,j]
    xMu = np.dot(V,Snew)
    if (np.linalg.norm(xMu, 2) == 0):       # Solve the system as best we can...
        print "***LotharAlternate NO SOLUTION***"
        Snew[0,0] = mu
        xMu = np.dot(V,Snew)
    xMu = np.dot(xMu,Sfull)
    xMu = np.dot(xMu,bt)
    return xMu

def solveLotharOptimal(A,b,x,normNoise):
    """
    Returns the solution for x based on Lothar's new method of solving these.
    """
    [U,S,V] = np.linalg.svd(A)
    V = V.T
    m,n   = np.shape(A)
    bt    = np.dot(U.T, b)
    mu,empty    = optimalMu(A,bt,x,normNoise)
    Snew  = sigmaNew(S,mu)
    Sfull = np.diag(S)
    for j in range(0,n):
        Snew[j,j] = 1. / Snew[j,j]
    xMu = np.dot(V,Snew)
    xMu = np.dot(xMu,Sfull)
    xMu = np.dot(xMu,bt)
    return xMu

def solveLothar(A,b,normNoise):
    """
    Returns the solution for x based on Lothar's new method of solving these.
    """
    [U,S,V] = np.linalg.svd(A)
    V = V.T
    m,n   = np.shape(A)
    bt    = np.dot(U.T, b)
    mu    = tikhonovMu(S,bt,normNoise)
    Snew  = sigmaNew(S,mu)
    Sfull = np.diag(S)
    for j in range(0,n):
        Snew[j,j] = 1. / Snew[j,j]
    xMu = np.dot(V,Snew)
    xMu = np.dot(xMu,Sfull)
    xMu = np.dot(xMu,bt)
    return xMu

def solveTSVDTikhonov(A,b,normNoise):
    """
    Returns the solution vector x_k corresponding to the Tikhonov regularaization parameter
    solved by setting sigma_k = 0 for all sigma_k < mu.
    """
    [U,S,V] = np.linalg.svd(A)
    V = V.T
    m,n = np.shape(A)
    mu = tikhonovMu(S,np.dot(U.T, b),normNoise)
    w = np.dot(U.T, b)
    y = np.zeros(n)
    for j in range(0,n):
        if S[j]**2 > mu:
            y[j] = w[j] / S[j]
    if (np.linalg.norm(y, 2) == 0):     # Solve the system as best we can...
        print "***TSVDTik NO SOLUTION***"
        y[0] = w[0] / mu
    return np.dot(V,y)

